/*
 * Purrpackage - http://purrpackage.sourceforge.net
 * Copyright (C) 2010 John Charles Roth
 *
 * Purrpackage is free software, licensed under the terms of the GNU GPL 
 * Version 2 or, at your option, any later version. You should have 
 * received a copy of the license with this file. See the above web address
 * for more information, or contact the Free Software Foundation, Boston, MA. 
 * It is distributed WITHOUT WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
package net.sourceforge.purrpackage.model;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import net.sourceforge.cobertura.coveragedata.CoverageData;

/**
 * Generic node in the coverage tree. For any node, the raw coverage data for
 * all tests ( {@link #getData()}) and for tests in the same package (
 * {@link #getSamePackageData()}) are summarized by the values of
 * {@link #getLineCounts()}, {@link #getBranchCounts()}, and
 * {@link #getElementCounts()} methods, the last being the sum of the first two.
 * 
 * A single {@link ProjectDataPair} is at the top of the tree, with
 * {@link PackageDataPair} children, which in turn has
 * {@link SourceFileDataPair} children, and so on with {@link ClassDataPair},
 * {@link MethodDataPair}, and {@link LineDataPair}. The generic types are made
 * concrete by each concrete subclass.
 * 
 * @param <DataType> type of {@link #getData()} and {@link #getSamePackageData()}
 * @param <ParentType> type of parent element of this node type.
 * @param <ChildType> type of child element of this node type
 */

public abstract class CoverageDataPair <DataType extends CoverageData, 
	ParentType extends CoverageDataPair<?,?,?>, ChildType extends CoverageDataPair<?,?,?>>{
	
	DataType data, samePackageData;
	CoverageSummaryView lineCounts, branchCounts, elementCounts;
	ParentType parent;	

	Map<Object,ChildType> children = new LinkedHashMap<Object, ChildType>();
		
	/** Coverage from running all tests */
	public DataType getData() {
		return data;
	}
	
	/** Coverage from running tests in the same package as represented by this node, if available. */
	public DataType getSamePackageData() {
		return samePackageData;
	}
	
	public ParentType getParent( ) {
		return parent;
	}

	public void addChild( ChildType value ) {
		children.put( value.getName(), value );
	}
	
	public ProjectDataPair getRoot() {
		return parent.getRoot();
	}

	/**
	 * Copies the children's keySet into a new List.
	 */
	public List<Object> getChildKeys() {
		return new ArrayList<Object>( children.keySet() );
	}

	public ChildType getChild( Object key ) {
		return children.get( key );
	}
	
	public List<Object[]> getGrandchildKeys () {
		ArrayList<Object[]> result = new ArrayList<Object[]>();
		for( Object key : getChildKeys() ) {
			for ( Object gkey : getChild( key ).getChildKeys() ) {
				result.add( new Object[] { key, gkey } );
			}
		}
		return result;
	}
	public List<Object[]> getGreatGrandchildKeys () {
		ArrayList<Object[]> result = new ArrayList<Object[]>();
		for( Object key : getChildKeys() ) {
			ChildType child = getChild( key );
			for ( Object gkey : child.getChildKeys() ) {
				for ( Object ggkey : child.getChild( gkey ).getChildKeys() ) {
					result.add( new Object[] { key, gkey, ggkey } );
				}
			}
		}
		return result;
	}
	
	/**
	 * @param keyOrKeys
	 *            If an array, treated as a path to descendent. If nonArray,
	 *            treated as {@link #getVanillaDescendent(Object)}. If
	 * @return the descendent, if any.
	 */

	public CoverageDataPair<?,?,?> getVanillaDescendent( Object keyOrKeys ) {
		if ( keyOrKeys.getClass().isArray() ) {
			return getVanillaNDeepDescendent( (Object[]) keyOrKeys );
		}
		else {
			return getChild(keyOrKeys);
		}
	}

	private CoverageDataPair<?,?,?> getVanillaNDeepDescendent( Object[] multiKey ) {
		CoverageDataPair<?,?,?> result = this;
		for( int i = 0; i < multiKey.length; i++ ) {
			if ( result != null ) {
				result = result.getChild(multiKey[i]);
			}
		}
		return result;
	}
	
	public CoverageDataPair( DataType data, DataType samePackageData, ParentType parent ) {
		this.data = data;
		this.samePackageData = samePackageData;
		this.parent = parent;
		lineCounts = new CoverageSummaryView( data.getNumberOfValidLines(),
				data.getNumberOfCoveredLines(), 
				samePackageData == null ? 0 : samePackageData.getNumberOfCoveredLines() );
		branchCounts = new CoverageSummaryView( data.getNumberOfValidBranches(),
				data.getNumberOfCoveredBranches(), 
				samePackageData == null ? 0 : samePackageData.getNumberOfCoveredBranches() );		
		elementCounts = new CoverageSummaryView( lineCounts.getValid() + branchCounts.getValid(),
				lineCounts.getCovered() + branchCounts.getCovered(),
				lineCounts.getSamePackageCovered() + branchCounts.getSamePackageCovered() );
	}
	
	public CoverageSummaryView getLineCounts() {
		return lineCounts;
	}
	public CoverageSummaryView getBranchCounts() {
		return branchCounts;
	}

	/** Combination of line and branch coverage */
	public CoverageSummaryView getElementCounts() {
		return elementCounts;
	}
	
	/**
	 * An object representing the natural ordering of nodes of this type.
	 * Perhaps by name, perhaps by a line number.
	 */
	public abstract Comparable<?> getOrderingObject();
	
	public Object getName() {
		return getOrderingObject();
	}
	
	public String getDisplayName() {
		String result = String.valueOf( getOrderingObject() );
		return result;
	}

	public boolean getHasBranches() {
		return getBranchCounts().getValid() > 0;
	}
	
	CoverageDataPairFactory getFactory() {
		return parent.getFactory();
	}
	

}
