/*
 * Purrpackage - http://purrpackage.sourceforge.net
 * Copyright (C) 2010 John Charles Roth
 *
 * Purrpackage is free software, licensed under the terms of the GNU GPL 
 * Version 2 or, at your option, any later version. You should have 
 * received a copy of the license with this file. See the above web address
 * for more information, or contact the Free Software Foundation, Boston, MA. 
 * It is distributed WITHOUT WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

package net.sourceforge.purrpackage.recording;

import java.io.File;
import java.io.IOException;

import net.sourceforge.cobertura.coveragedata.ProjectData;
import net.sourceforge.cobertura.util.ConfigurationUtil;

/**
 * The simple guts of PurrPackage recording.
 * 
 * This static singleton receives messages (via {@link #inPackage(String)} about
 * which package a test about to be run belongs to. When the package changes, it
 * saves off a specially named copy of the coverage data file, and adjusts the
 * purrpackage.db file to track the change.
 * 
 * @see PerPackageCoverageListener
 * @see PerPackageCoverageJUnitFormatter
 */

public class CoberturaManager {

	/** Key to {@link System#getProperties()} that stores the file Cobertura is using.*/
	private static final String COB_FILE_PROP = "net.sourceforge.cobertura.datafile";
	static final String UNUSED_PACKAGE_NAME = "net.sourceforge.purrpackage.recording.unused.package";

	CoberturaManager() {
	}

	/**
	 * Call this just before running a test in the given package.
	 */
	public void inPackage( String packageName ) {
		if (!packageName.equals( getCurrentPackage())) {
			setCurrentPackage( packageName );
			rollCobertura();
		}		
	}

	String currentPackage;
	PropDb db;

	public String getCurrentPackage() {
		return currentPackage;
	}

	public void setCurrentPackage(String currentPackage) {
		this.currentPackage = currentPackage;
	}

	ConfigurationUtil configurationUtil = new ConfigurationUtil();
	
	String coverageFileName() {
		return configurationUtil.getDatafile();
	}
	
	void rollCobertura() {
		String cobFileName = coverageFileName();
		if (cobFileName == null) {
			throw new RuntimeException("Cobertura data file is not defined; use System property "
					+ COB_FILE_PROP + " or a cobertura.properties file in the classpath." );
		}
		try {
			ProjectData.saveGlobalProjectData();
			File cser = new File(cobFileName);
			if (!cser.exists()) {
				throw new RuntimeException("No cobertura file at "
						+ cobFileName);
			}
			db = new PropDb();
			db.read(cser.getParentFile());
			copyFileToPerPackageArchive(cser);
			int count = 0;
			if (db.getCount() != null) {
				count = db.getCount();
			}
			db.setCount(count + 1);
			db.setPackageName(currentPackage);
			db.write(cser.getParentFile());
		} catch (Throwable t) {
			throw new RuntimeException("Failed to roll cobertura file: " + t, t);
		}
	}
	
	String nextFileName() throws IOException {
		int newCount = 0;
		if ( db.getCount() != null ) {
			newCount = db.getCount();
		}
		String packageName = db.getPackageName();
		if ( packageName == null ) {
			packageName = UNUSED_PACKAGE_NAME;
		}
		String newFilename = packageName + "-" + newCount + "-" + "cobertura.ser";
		return newFilename;
	}

	void copyFileToPerPackageArchive(File cobSerFile) throws IOException {
		String fileName = nextFileName();
		File dir = cobSerFile.getParentFile();
		Copier c = new Copier();
		c.copyFile( cobSerFile, new File( dir, fileName ));
	}

	private static CoberturaManager instance;

	public synchronized static CoberturaManager getInstance() {
		if ( instance == null ) {
			instance = new CoberturaManager();
		}
		return instance;
	}
	
	static void setInstanceForTesting( CoberturaManager x ) {
		instance = x;
	}

}
